/********************************** (C) COPYRIGHT *******************************
 * File Name          : main.c
 * Author             : WCH
 * Version            : V1.0.0
 * Date               : 2023/12/12
 * Description        : Main program body.
*********************************************************************************
* Copyright (c) 2023 Nanjing Qinheng Microelectronics Co., Ltd.
* Attention: This software (modified or not) and binary are used for 
* microcontroller manufactured by Nanjing Qinheng Microelectronics.
*******************************************************************************/
/*
 *@Note
 This example demonstrates the use of LIN through USART2.
     Hardware connection:PA2 -- Tx
                         PA3 -- Rx

*/

#include "debug.h"
#include "string.h"
/* Global define */
#define HOST_MODE   0
#define SLAVE_MODE   1

//#define LIN_MODE   HOST_MODE
#define LIN_MODE   SLAVE_MODE

#define LIN_BAUDRATE (19200u)
#define LIN_SYNC_LEN (1u)
#define LIN_ID_LEN (1u)
#define LIN_DATA_LEN (8u)
#define LIN_CHECKSUM_LEN (1u)
#define LIN_PACK_LEN (LIN_SYNC_LEN+LIN_ID_LEN+LIN_DATA_LEN+LIN_CHECKSUM_LEN)

#define LIN_SYNC_DATA (0x55u)
#define LIN_MSATER_ID (0x21)
#define LIN_SLAVE_ID (0x22)

/* Global typedef */
typedef struct {
    uint8_t sync;
    uint8_t ID;
    uint8_t dataBuf[LIN_DATA_LEN];
    uint8_t checkSum;
} LIN_Data;

/* Global Variable */
LIN_Data lin_pack_recv = {0};
LIN_Data lin_pack_send = {0};

uint8_t lin_master_msg[LIN_DATA_LEN] = { 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18 };
uint8_t lin_slave_msg[LIN_DATA_LEN] = { 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28 };

uint8_t lin_recv_buffer[LIN_PACK_LEN] = { 0 };

volatile uint8_t Rx_Break_Flag = 0;
volatile uint8_t Rx_Valid_Data = 0;
volatile uint8_t Rx_Slave_ID = 0;
volatile uint8_t Rx_Timeout = 0;
volatile uint8_t Rx_Cnt = 0;

void USART2_IRQHandler(void) __attribute__((interrupt("WCH-Interrupt-fast")));
void TIM3_IRQHandler(void) __attribute__((interrupt("WCH-Interrupt-fast")));

/*********************************************************************
 * @fn      USART2_LIN_Init
 *
 * @brief   Initializes the USART2 peripheral.
 *
 * @return  none
 */
void USART2_LIN_Init(uint32_t baudrate) {
    GPIO_InitTypeDef GPIO_InitStructure = {0};
    USART_InitTypeDef USART_InitStructure = {0};
    NVIC_InitTypeDef NVIC_InitStructure = {0};

    RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOA, ENABLE);
    RCC_APB1PeriphClockCmd(RCC_APB1Periph_USART2, ENABLE);

    //USART2_TX   PA.2
    GPIO_InitStructure.GPIO_Pin = GPIO_Pin_2;
    GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
    GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF_PP;
    GPIO_Init(GPIOA, &GPIO_InitStructure);

    //USART2_RX   PA.3
    GPIO_InitStructure.GPIO_Pin = GPIO_Pin_3;
    GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN_FLOATING;
    GPIO_Init(GPIOA, &GPIO_InitStructure);

    USART_InitStructure.USART_BaudRate = baudrate;
    USART_InitStructure.USART_WordLength = USART_WordLength_8b;
    USART_InitStructure.USART_StopBits = USART_StopBits_1;
    USART_InitStructure.USART_Parity = USART_Parity_No;
    USART_InitStructure.USART_HardwareFlowControl = USART_HardwareFlowControl_None;
    USART_InitStructure.USART_Mode = USART_Mode_Rx | USART_Mode_Tx;

    USART_Init(USART2, &USART_InitStructure);

    NVIC_InitStructure.NVIC_IRQChannel = USART2_IRQn;
    NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0;
    NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;
    NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
    NVIC_Init(&NVIC_InitStructure);

    USART_ITConfig(USART2, USART_IT_RXNE, ENABLE);
    USART_ITConfig(USART2, USART_IT_LBD, ENABLE);

    USART_LINBreakDetectLengthConfig(USART2, USART_LINBreakDetectLength_11b);

    USART_LINCmd(USART2, ENABLE);

    USART_Cmd(USART2, ENABLE);
}



/*********************************************************************
 * @fn      TIM3_init
 *
 * @brief   Initializes the TIM3 peripheral.
 *
 * @return  none
 */
void TIM3_init()
{

    TIM_TimeBaseInitTypeDef TIM_TimeBaseStructure = {0};
    NVIC_InitTypeDef NVIC_InitStructure = {0};
    RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM3, ENABLE);
    TIM_Cmd(TIM3, DISABLE);

    TIM_TimeBaseStructure.TIM_Period        = 1000 - 1;
    TIM_TimeBaseStructure.TIM_Prescaler     = SystemCoreClock / 100000 - 1;
    TIM_TimeBaseStructure.TIM_ClockDivision = TIM_CKD_DIV1;
    TIM_TimeBaseStructure.TIM_CounterMode   = TIM_CounterMode_Up;
    TIM_TimeBaseInit(TIM3, &TIM_TimeBaseStructure);
    TIM_SetCounter(TIM3, 0);

    TIM_ClearFlag(TIM3, TIM_FLAG_Update);
    TIM_ITConfig(TIM3, TIM_IT_Update, ENABLE);

    NVIC_InitStructure.NVIC_IRQChannel                   = TIM3_IRQn;
    NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 1;
    NVIC_InitStructure.NVIC_IRQChannelSubPriority        = 1;
    NVIC_InitStructure.NVIC_IRQChannelCmd                = ENABLE;
    NVIC_Init(&NVIC_InitStructure);

    TIM_Cmd(TIM3, DISABLE);
}

/*********************************************************************
 * @fn      LIN_GetId
 *
 * @brief   Get ID.
 *
 * @param   ID - the lin ID.
 *
 * @return  PID
 */
uint8_t LIN_GetId(uint8_t ID) {
    static const uint8_t idParityTable[64] =
    {
            0x80, 0xC1, 0x42, 0x03, 0xC4, 0x85, 0x06, 0x47,
            0x08, 0x49, 0xCA, 0x8B, 0x4C, 0x0D, 0x8E, 0xCF,
            0x50, 0x11, 0x92, 0xD3, 0x14, 0x55, 0xD6, 0x97,
            0xD8, 0x99, 0x1A, 0x5B, 0x9C, 0xDD, 0x5E, 0x1F,
            0x20, 0x61, 0xE2, 0xA3, 0x64, 0x25, 0xA6, 0xE7,
            0xA8, 0xE9, 0x6A, 0x2B, 0xEC, 0xAD, 0x2E, 0x6F,
            0xF0, 0xB1, 0x32, 0x73, 0xB4, 0xF5, 0x76, 0x37,
            0x78, 0x39, 0xBA, 0xFB, 0x3C, 0x7D, 0xFE, 0xBF,
    };
    return idParityTable[ID];
}


/*********************************************************************
 * @fn      LIN_CheckSum
 *
 * @brief   Generate checksum.
 *
 * @param   data - lin data.
 *          frameID - the lin ID.
 *          len - data length.
 *
 * @return  checksum
 */
uint8_t LIN_CheckSum(uint8_t *data, uint8_t frameID, uint16_t len) {
    uint16_t tmp = frameID;
	uint8_t result = 0x0;
	uint8_t i;
    if(((frameID & 0x3f) == 62) || ((frameID & 0x3f) == 63))
    {
        tmp = 0;
    }

    for (i = 0; i < len; i++) {
        tmp += data[i];
        if (tmp >= 256) {
            tmp -= 255;
        }
    }
    result = ~tmp & 0xFF;
    return result;
}

/*********************************************************************
 * @fn      LIN_CreatePack
 *
 * @brief   Create a lin packet.
 *
 * @param   lin_pack - the result.
 *          data - lin data.
 *          ID - the lin ID.
 *
 * @return  .
 */
void LIN_CreatePack(LIN_Data *lin_pack, uint8_t *msg, uint8_t ID) {
    lin_pack->sync = LIN_SYNC_DATA;
    lin_pack->ID = LIN_GetId(ID & 0x3f);
    memcpy(lin_pack->dataBuf, msg, LIN_DATA_LEN);
    lin_pack->checkSum = LIN_CheckSum(lin_pack->dataBuf, lin_pack->ID, LIN_DATA_LEN);
}


/*********************************************************************
 * @fn      LIN_MasterSendHead
 *
 * @brief   Send a lin header with the specified id.
 *
 * @param   ID - the lin ID.
 *
 * @return  .
 */
void LIN_MasterSendHead(uint8_t ID)
{
    while (USART_GetFlagStatus(USART2, USART_FLAG_TC) == RESET);
    USART_SendBreak(USART2);

    // wait for break
    while(USART2->CTLR1 & USART_CTLR1_SBK);
    Delay_Us((1000*1000)/LIN_BAUDRATE);
    //   1 * ((1/baudrate)*1000*1000)
    //   1 * ((1/19200)*1000*1000) = 52 us

    USART_SendData(USART2, LIN_SYNC_DATA);
    while (USART_GetFlagStatus(USART2, USART_FLAG_TC) == RESET);
    USART_SendData(USART2, ID);
}


/*********************************************************************
 * @fn      LIN_SendResponse
 *
 * @brief   Send a response.
 *
 * @param   lin_pack - the lin data packet.
 *
 * @return  .
 */
void LIN_SendResponse(LIN_Data *lin_pack)
{
    uint16_t i = 0;
    while(i < LIN_DATA_LEN){
        while (USART_GetFlagStatus(USART2, USART_FLAG_TC) == RESET);
        USART_SendData(USART2, lin_pack->dataBuf[i]);
        i++;
    }
    while (USART_GetFlagStatus(USART2, USART_FLAG_TC) == RESET);
    USART_SendData(USART2, lin_pack->checkSum);
    while (USART_GetFlagStatus(USART2, USART_FLAG_TC) == RESET);
}



/*********************************************************************
 * @fn      LIN_MasterSendMsg
 *
 * @brief   Send a lin header and lin response.
 *
 * @param   lin_pack - the lin data packet.
 *
 * @return  .
 */
void LIN_MasterSendMsg(LIN_Data *lin_pack)
{
    LIN_MasterSendHead(lin_pack->ID);
    LIN_SendResponse(lin_pack);
}

/*********************************************************************
 * @fn      LIN_MasterSendNotify
 *
 * @brief   The master sends the specified id header and starts the timer.
 *
 * @param   ID - the lin ID.
 *
 * @return  .
 */
void LIN_MasterSendNotify(uint8_t ID)
{
    ID = LIN_GetId(ID & 0x3f);
    LIN_MasterSendHead(ID);
    Rx_Timeout = 0 ;
    TIM_Cmd(TIM3, ENABLE);
}

/*********************************************************************
 * @fn      LIN_RxBufferPorcesses
 *
 * @brief   Convert buffer to packet.
 *
 * @param   lin_pack - the lin data packet.
 *          buffer - the lin data buffer.
 *
 * @return  .
 */
void LIN_RxBufferPorcesses(LIN_Data *lin_pack, uint8_t *buffer){
    uint8_t i = 0;
    uint8_t j;
    lin_pack->sync = buffer[i];
    i += LIN_SYNC_LEN;
    lin_pack->ID = buffer[i];
    i += LIN_ID_LEN;
    for (j = 0; j < LIN_DATA_LEN; j++) {
        lin_pack->dataBuf[j] =  buffer[i];
        i++;
    }
    lin_pack->checkSum = buffer[i];
}



/*********************************************************************
 * @fn      master_process
 *
 * @brief   Master processes Lin messages.
 *
 * @return  none
 */
void master_process()
{
    uint8_t checkSum;
    if(Rx_Timeout&&(!Rx_Valid_Data)){
           Rx_Timeout = 0;
           printf("Rx_Timeout,cannot read data !!\n");
    }
    if (Rx_Valid_Data == 1) {
        TIM_Cmd(TIM3, DISABLE);
        TIM_SetCounter(TIM3, 0);
        LIN_RxBufferPorcesses(&lin_pack_recv,lin_recv_buffer);
        // lin echo
        if((lin_pack_recv.ID & 0x3f) == LIN_MSATER_ID)
        {
            printf("LIN ECHO \n");
            Rx_Valid_Data = 0;
            return;
        }
        printf("sync %x \n", lin_pack_recv.sync);
        printf("ID %x \n", (lin_pack_recv.ID & 0x3f));
        if (LIN_GetId(lin_pack_recv.ID & 0x3f) != lin_pack_recv.ID) {
            printf("Incorrect ID\n");
        }
        for (int i = 0; i < LIN_DATA_LEN; ++i) {
            printf("[%d] %02x\n", i, lin_pack_recv.dataBuf[i]);
        }
        checkSum = LIN_CheckSum(lin_pack_recv.dataBuf, lin_pack_recv.ID, LIN_DATA_LEN);
        printf("checkSum %x \n", lin_pack_recv.checkSum);
        if(checkSum != lin_pack_recv.checkSum)
        {
            printf("the checkSum should be %x \n", checkSum);
        }
        Rx_Valid_Data = 0;
    }
}


/*********************************************************************
 * @fn      slave_procrss
 *
 * @brief   Slave processes Lin messages.
 *
 * @return  none
 */
void slave_procrss()
{
    uint8_t checkSum;
    if(Rx_Timeout&&(!Rx_Valid_Data)){
           Rx_Timeout = 0;
           printf("Rx_Timeout,cannot read data !!\n");
    }
    if (Rx_Valid_Data == 1) {
        TIM_Cmd(TIM3, DISABLE);
        TIM_SetCounter(TIM3, 0);
        LIN_RxBufferPorcesses(&lin_pack_recv,lin_recv_buffer);
        printf("sync %x \n", lin_pack_recv.sync);
        if(lin_pack_recv.sync != LIN_SYNC_DATA){
            printf("sync fail\n ");
        }
        printf("ID %x \n", (lin_pack_recv.ID & 0x3f));
        if (LIN_GetId(lin_pack_recv.ID & 0x3f) != lin_pack_recv.ID) {
            printf("Incorrect ID\n");
        }
        for (int i = 0; i < LIN_DATA_LEN; ++i) {
            printf("[%d] %02x\n", i, lin_pack_recv.dataBuf[i]);
        }
        checkSum = LIN_CheckSum(lin_pack_recv.dataBuf, lin_pack_recv.ID, LIN_DATA_LEN);
        printf("checkSum %x \n", lin_pack_recv.checkSum);
        if(checkSum != lin_pack_recv.checkSum)
        {
            printf("the checkSum should be %x \n", checkSum);
        }

        Rx_Valid_Data = 0;
    }
    // recive the slave ID, send lin response
    else if(Rx_Slave_ID){
        Rx_Slave_ID = 0;
        TIM_Cmd(TIM3, DISABLE);
        TIM_SetCounter(TIM3, 0);
        LIN_RxBufferPorcesses(&lin_pack_recv,lin_recv_buffer);
        LIN_CreatePack(&lin_pack_send, lin_slave_msg, lin_pack_recv.ID);
        LIN_SendResponse(&lin_pack_send);
        printf("send response \n");

    }
}

/*********************************************************************
 * @fn      main
 *
 * @brief   Main program.
 *
 * @return  none
 */
int main(void) {
    uint16_t i = 0;
    NVIC_PriorityGroupConfig(NVIC_PriorityGroup_2);
    Delay_Init();
    USART_Printf_Init(115200);
    printf("SystemClk:%d\r\n", SystemCoreClock);

    printf("USART LIN TEST\r\n");
    TIM3_init();
#if LIN_MODE == HOST_MODE
    printf("master mode\r\n");
    LIN_CreatePack(&lin_pack_send, lin_master_msg, LIN_MSATER_ID);
#else
    printf("Slave mode\r\n");

#endif
    USART2_LIN_Init(LIN_BAUDRATE);

    while (1)
    {
        i++;
#if LIN_MODE == HOST_MODE
        Delay_Ms(1);
        if(i == 500) {
            LIN_MasterSendMsg(&lin_pack_send);
        }
        if(i > 550) {
            LIN_MasterSendNotify(LIN_SLAVE_ID);
            i = 0;
        }

        master_process();

#else
        slave_procrss();
#endif
    }

}


/*********************************************************************
 * @fn      TIM3_IRQHandler
 *
 * @brief   This function handles TIM3 global interrupt request.
 *
 * @return  none
 */
void TIM3_IRQHandler(void) {
    if (TIM_GetITStatus(TIM3, TIM_IT_Update)) {
        TIM_Cmd(TIM3, DISABLE);
        TIM_ClearITPendingBit(TIM3, TIM_IT_Update);
        Rx_Timeout = 1;
        if(Rx_Cnt != LIN_PACK_LEN){
            Rx_Break_Flag = 0;
            Rx_Valid_Data = 0;
            Rx_Cnt = 0;
        }
    }

}

/*********************************************************************
 * @fn      USART2_IRQHandler
 *
 * @brief   This function handles USART2 global interrupt request.
 *
 * @return  none
 */
void USART2_IRQHandler(void) {
    uint8_t res;
    if (USART_GetITStatus(USART2, USART_IT_LBD) != RESET) {
        Rx_Break_Flag = 1;
        USART_ClearITPendingBit(USART2, USART_IT_LBD);
        TIM_Cmd(TIM3, ENABLE );
    }
    else if (USART_GetITStatus(USART2, USART_IT_RXNE) != RESET) {
        res = USART_ReceiveData(USART2);
#if LIN_MODE == HOST_MODE
        // master receive data
        if (Rx_Break_Flag) {
            lin_recv_buffer[Rx_Cnt] = res;
            Rx_Cnt++;
            if (Rx_Cnt == LIN_PACK_LEN) {
                Rx_Break_Flag = 0;
                Rx_Valid_Data = 1;
                Rx_Cnt = 0;
            }
        } else {
            Rx_Cnt = 0;
        }

#else
        // slave receive data
        if (Rx_Break_Flag) {
            lin_recv_buffer[Rx_Cnt] = res;
            Rx_Cnt++;
            if (Rx_Cnt == LIN_SYNC_LEN+LIN_ID_LEN) {
                if((res & 0x3f ) == LIN_SLAVE_ID ){
                    Rx_Break_Flag = 0;
                    Rx_Cnt = 0;
                    Rx_Slave_ID = 1;
                }
            } else if (Rx_Cnt == LIN_PACK_LEN) {
                Rx_Break_Flag = 0;
                Rx_Valid_Data = 1;
                Rx_Cnt = 0;
            }
        } else {
            Rx_Cnt = 0;
        }
#endif
    }
}
